﻿
Module Module_CNC

    ' ------------------------------------------ enums
    Friend Enum CNC_FeedSpeedLockedTypes
        AllFromGcode
        FeedLocked
        SpeedLocked
        AllLocked
    End Enum
    Friend Enum CNC_LookAheadTypes
        None
        LookAhead
        LookAheadHiPrecision
        ExactStopAllMotors
    End Enum

    Friend Enum CalibrationSteps
        None
        X1
        X2
        X3
        Y1
        Y2
        Y3
        Z1
        Z2
        Z3
        A1
        A2
        A3
        B1
        B2
        B3
    End Enum
    Friend CNC_CalibrationStep As CalibrationSteps

    ' ------------------------------------------ gcode
    Friend CNC_GcodeParams As GCODE_PARAMS

    ' ------------------------------------------ slots
    Friend SlotDestX As Int32
    Friend SlotDX As Int32
    Friend SlotDestY As Int32
    Friend SlotDY As Int32
    Friend SlotDestZ As Int32
    Friend SlotDZ As Int32
    Friend SlotDestA As Int32
    Friend SlotDA As Int32
    Friend SlotDestB As Int32
    Friend SlotDB As Int32
    Friend SlotSpindleOnOff As Int32
    Friend SlotSpindleSpeed As Int32
    Friend SlotToolNumber As Int32
    Friend SlotCooling1 As Int32
    Friend SlotCooling2 As Int32
    Friend SlotCooling3 As Int32
    Friend SlotCooling4 As Int32
    Friend SlotPalletClamp As Int32
    Friend SlotSpindleOrientation As Int32
    Friend SlotGearSpeed As Int32
    Friend SlotMainEnable As Int32
    Friend SlotEmergencyInput As Int32
    Friend SlotLimitsInput As Int32
    Friend SlotCalibrationInput As Int32
    ' ------------------------------------------ working positions
    Friend CNC_Dest As Vec5
    Friend CNC_Tip As Vec5
    Friend CNC_Home_Z As Double
    ' ------------------------------------------ hardware tip (different if IN-OUT disabled)
    Friend CNC_HardwareTip As Vec5
    Friend CNC_HardwareHome_Z As Double
    ' ------------------------------------------ 
    Friend CNC_ToolStop As Boolean
    Friend CNC_TestLowZ As Boolean
    Friend CNC_FeedMode As Feed_Modes
    Friend CNC_Rapid As Int32
    Friend CNC_Feed As Double
    Friend CNC_Speed As Double
    Friend CNC_FeedSpeedLockedType As CNC_FeedSpeedLockedTypes
    Friend CNC_LookAheadType As CNC_LookAheadTypes
    Friend CNC_InOutEnabled As Boolean
    ' ------------------------------------------ 
    Friend CNC_TimeCoefficient As Double
    Friend CNC_MaxError As Double
    Friend CNC_JogStep As Double
    ' ------------------------------------------
    Friend CNC_ProbeThickness As Double

    ' ------------------------------------------ 
    Friend CNC_GcodeRunning As Boolean = False
    Friend CNC_LineInExecution As Int32 = -1
    Friend CNC_LineToBeExecuted As Int32 = -1
    Friend CNC_WarningZeta As Double = 1
    ' ------------------------------------------ 
    Friend CNC_CalXyabSearchSpeed As Int32
    Friend CNC_CalXyabReturnSpeed As Int32
    Friend CNC_CalXyabMaxTravel As Int32
    Friend CNC_CalXyabCompensation As Single
    Friend CNC_CalZSearchSpeed As Int32
    Friend CNC_CalZReturnSpeed As Int32
    Friend CNC_CalZMaxTravel As Int32
    Friend CNC_CalZCompensation As Single
    Friend CNC_CalFinalClearance As Single = 1 ' final movement fixed = 1 mm

    ' ==================================================================================================
    '   INIT
    ' ==================================================================================================
    Friend Sub CNC_Init(ByRef _ToolpathPic As PictureBox)
        ToolpathPic = _ToolpathPic
        CNC_SetFeedMode(Feed_Modes.None)
    End Sub

    'Friend Sub CNC_InitGcodeParams()
    '    CNC_GcodeParams.FeedMode = Feed_Modes.None
    '    CNC_GcodeParams.Coord = CNC_Tip
    'End Sub

    Friend Sub CNC_RestoreOutputSignals()
        With GCodeParsedLines(CNC_LineInExecution)
            If Not CNC_FeedSpeedLockedType = CNC_FeedSpeedLockedTypes.AllLocked And _
               Not CNC_FeedSpeedLockedType = CNC_FeedSpeedLockedTypes.SpeedLocked Then
                CNC_SetSpeed(.Speed)
            End If
            If GCodeParsedLines(CNC_LineInExecution).FeedMode = Feed_Modes.Work Then
                Slots.WriteSlot(SlotSpindleOnOff, 1000)
            End If
            CNC_SetTool(.ToolNumber)
            CNC_SetCooling(GCodeParsedLines(CNC_LineInExecution))
            CNC_SetSpindleOrientation(.SpindleOrientation)
            CNC_SetGearSpeed(.GearSpeed)
            CNC_SetPalletClamp(.PalletClamp)
        End With
    End Sub

    Friend Sub CNC_SetTimeCoefficient(ByVal TimerSpeedMillisec As Double)
        ' ----------------------------------------------- correction  60 seconds per minute
        CNC_TimeCoefficient = TimerSpeedMillisec / 1000 / 60
    End Sub


    ' ==================================================================================================
    '   IN OUT - ALL FUNCTIONS ACCESSING IN-OUT ARE HERE
    ' ==================================================================================================
    Friend Sub CNC_SetSlots(ByVal FirstSlot As Int32)
        ' ------------------------------------------------------ Motors
        SlotDestX = FirstSlot       'Slot 1 (if first slot = 1)
        SlotDX = FirstSlot + 1      'Slot 2
        SlotDestY = FirstSlot + 2   'Slot 3
        SlotDY = FirstSlot + 3      'Slot 4
        SlotDestZ = FirstSlot + 4   'Slot 5
        SlotDZ = FirstSlot + 5      'Slot 6
        SlotDestA = FirstSlot + 6   'Slot 7
        SlotDA = FirstSlot + 7      'Slot 8
        SlotDestB = FirstSlot + 8   'Slot 9
        SlotDB = FirstSlot + 9      'Slot 10
        ' ------------------------------------------------------ Outputs
        SlotSpindleOnOff = FirstSlot + 10       'Slot 11
        SlotSpindleSpeed = FirstSlot + 11       'Slot 12
        SlotToolNumber = FirstSlot + 12         'Slot 13
        SlotCooling1 = FirstSlot + 13           'Slot 14
        SlotCooling2 = FirstSlot + 14           'Slot 15
        SlotCooling3 = FirstSlot + 15           'Slot 16
        SlotCooling4 = FirstSlot + 16           'Slot 17
        SlotPalletClamp = FirstSlot + 17        'Slot 18
        SlotSpindleOrientation = FirstSlot + 18 'Slot 19
        SlotGearSpeed = FirstSlot + 19          'Slot 20
        SlotMainEnable = FirstSlot + 20         'Slot 21
        'Free 1 = FirstSlot + 21                'Slot 22
        'Free 2 = FirstSlot + 22                'Slot 23
        'Free 3 = FirstSlot + 23                'Slot 24
        ' ------------------------------------------------------ Inputs
        SlotEmergencyInput = FirstSlot + 24     'Slot 25
        SlotLimitsInput = FirstSlot + 25        'Slot 26
        SlotCalibrationInput = FirstSlot + 26   'Slot 27
    End Sub

    Friend Sub CNC_WriteSlotValues()
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotDestX, CSng(CNC_Tip.x))
        Slots.WriteSlot(SlotDestY, CSng(CNC_Tip.y))
        Slots.WriteSlot(SlotDestZ, CSng(CNC_Tip.z))
        Slots.WriteSlot(SlotDestA, CSng(CNC_Tip.a))
        Slots.WriteSlot(SlotDestB, CSng(CNC_Tip.b))
    End Sub

    Friend Sub CNC_ResetAllAxis()
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotDestX, NAN_Reset)
        Slots.WriteSlot(SlotDestY, NAN_Reset)
        Slots.WriteSlot(SlotDestZ, NAN_Reset)
        Slots.WriteSlot(SlotDestA, NAN_Reset)
        Slots.WriteSlot(SlotDestB, NAN_Reset)
        Threading.Thread.Sleep(100)
    End Sub

    Friend Sub CNC_EnsureZeroToDeltaSlots()
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotDX, 0)
        Slots.WriteSlot(SlotDY, 0)
        Slots.WriteSlot(SlotDZ, 0)
        Slots.WriteSlot(SlotDA, 0)
        Slots.WriteSlot(SlotDB, 0)
    End Sub

    Friend Sub CNC_ResetHardwareTo(ByVal axis As String, ByVal coord As Double)
        Dim slot As Int32
        Select Case axis(axis.Length - 1)
            Case "X"c : slot = SlotDestX : CNC_Dest.x = coord : CNC_Tip.x = coord
            Case "Y"c : slot = SlotDestY : CNC_Dest.y = coord : CNC_Tip.y = coord
            Case "Z"c : slot = SlotDestZ : CNC_Dest.z = coord : CNC_Tip.z = coord
            Case "A"c : slot = SlotDestA : CNC_Dest.a = coord : CNC_Tip.a = coord
            Case "B"c : slot = SlotDestB : CNC_Dest.b = coord : CNC_Tip.b = coord
        End Select
        If CNC_InOutEnabled Then
            Slots.WriteSlot(slot, NAN_Reset)
            Threading.Thread.Sleep(150)
            Slots.WriteSlot(slot, CSng(coord))
        Else
            Threading.Thread.Sleep(150)
        End If
        GraphicThread_DrawToolpathImage(True)
    End Sub

    Friend Sub CNC_SetFeedMode(ByVal f As Feed_Modes)
        CNC_FeedMode = f
        CNC_GcodeParams.FeedMode = f
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotSpindleOnOff, If(CNC_FeedMode = Feed_Modes.Work, 1000, 0))
    End Sub

    Friend Sub CNC_SetRapid(ByVal rapid As Int32)
        If rapid < 10 Then rapid = 10
        CNC_Rapid = rapid
    End Sub

    Friend Sub CNC_SetFeed(ByVal feed As Int32)
        If feed < 10 Then feed = 10
        CNC_Feed = feed
    End Sub

    Friend Sub CNC_SetSpeed(ByVal speed As Int32)
        CNC_Speed = speed
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotSpindleSpeed, speed)
    End Sub

    Friend Sub CNC_SetTool(ByVal ToolNumber As Int32)
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotToolNumber, ToolNumber)
    End Sub

    Friend Sub CNC_SetCooling(ByVal gcp As GCODE_PARAMS)
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotCooling1, gcp.Cooling1)
        Slots.WriteSlot(SlotCooling2, gcp.Cooling2)
        Slots.WriteSlot(SlotCooling3, gcp.Cooling3)
        Slots.WriteSlot(SlotCooling4, gcp.Cooling4)
    End Sub

    Friend Sub CNC_ResetCoolings()
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotCooling1, 0)
        Slots.WriteSlot(SlotCooling2, 0)
        Slots.WriteSlot(SlotCooling3, 0)
        Slots.WriteSlot(SlotCooling4, 0)
    End Sub

    Friend Sub CNC_SetSpindleOrientation(ByVal SpindleOrientation As Single)
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotSpindleOrientation, SpindleOrientation)
    End Sub

    Friend Sub CNC_SetGearSpeed(ByVal GearSpeed As Int32)
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotGearSpeed, GearSpeed)
    End Sub

    Friend Sub CNC_SetPalletClamp(ByVal PalletClamp As Int32)
        If Not CNC_InOutEnabled Then Return
        Slots.WriteSlot(SlotPalletClamp, PalletClamp)
    End Sub

    Friend Sub CNC_SetMainEnableOutput(ByVal enable As Boolean)
        If Not CNC_InOutEnabled Then Return
        If enable Then
            Slots.WriteSlot(SlotMainEnable, 1000)
        Else
            Slots.WriteSlot(SlotMainEnable, 0)
        End If
    End Sub

    Friend Function CNC_MaxDelta() As Single
        If Not CNC_InOutEnabled Then Return 0
        Dim vmax As Single = 0
        Dim v As Single
        v = Math.Abs(Slots.ReadSlot_NoNan(SlotDX))
        If v > vmax Then vmax = v
        v = Math.Abs(Slots.ReadSlot_NoNan(SlotDY))
        If v > vmax Then vmax = v
        v = Math.Abs(Slots.ReadSlot_NoNan(SlotDZ))
        If v > vmax Then vmax = v
        v = Math.Abs(Slots.ReadSlot_NoNan(SlotDA))
        If v > vmax Then vmax = v
        v = Math.Abs(Slots.ReadSlot_NoNan(SlotDB))
        If v > vmax Then vmax = v
        Return vmax
    End Function

    Friend Function CNC_TestEmergencyInput() As Boolean
        If Not CNC_InOutEnabled Then Return False
        Return Slots.ReadSlot_NoNan(SlotEmergencyInput) > 500
    End Function

    Friend Function CNC_TestLimitInputs() As Boolean
        If Not CNC_InOutEnabled Then Return False
        If CNC_CalibrationStep <> CalibrationSteps.None Then Return False
        Return Slots.ReadSlot_NoNan(SlotLimitsInput) > 500
    End Function

    Friend Function CNC_TestCalibrationInputs() As Boolean
        If Not CNC_InOutEnabled Then Return True
        Return Slots.ReadSlot_NoNan(SlotLimitsInput) > 500 Or _
               Slots.ReadSlot_NoNan(SlotCalibrationInput) < 500
    End Function

    Friend Sub CNC_EnsureSafeValuesToInputSlots()
        Slots.WriteSlot(SlotEmergencyInput, 0)
        Slots.WriteSlot(SlotLimitsInput, 0)
        Slots.WriteSlot(SlotCalibrationInput, 1000)
    End Sub

    ' ==================================================================================================
    '   JOG
    ' ==================================================================================================
    Friend Sub SetDestinationFromPboxXY(ByVal x As Int32, ByVal y As Int32)
        CNC_Dest = CNC_DestFromXY(x, y)
    End Sub
    Private Sub JogX(ByVal delta As Double)
        CNC_Dest.x += delta
    End Sub
    Private Sub JogY(ByVal delta As Double)
        CNC_Dest.y += delta
    End Sub
    Private Sub JogZ(ByVal delta As Double)
        CNC_Dest.z += delta
    End Sub
    Private Sub JogA(ByVal delta As Double)
        CNC_Dest.a += delta
    End Sub
    Private Sub JogB(ByVal delta As Double)
        CNC_Dest.b += delta
    End Sub

    Friend Sub CNC_Jog(ByVal key As Int32, ByVal mm As Double)
        Select Case key
            Case Keys.Left
                JogX(-mm)
            Case Keys.Right
                JogX(mm)
            Case Keys.Up
                JogY(mm)
            Case Keys.Down
                JogY(-mm)
            Case Keys.PageUp, Keys.X
                JogZ(mm)
            Case Keys.PageDown, Keys.Z
                JogZ(-mm)
            Case Keys.A
                JogA(-mm)
            Case Keys.S
                JogA(mm)
            Case Keys.B
                JogB(-mm)
            Case Keys.N
                JogB(mm)
        End Select
    End Sub

End Module
